__author__ = 'Lukas Leufen'
__date__ = '2020-11-05'

from mlair.data_handler.data_handler_single_station import DataHandlerSingleStation
from mlair.data_handler.data_handler_with_filter import DataHandlerKzFilterSingleStation, \
    DataHandlerFirFilterSingleStation, DataHandlerFilterSingleStation, DataHandlerClimateFirFilterSingleStation
from mlair.data_handler import DefaultDataHandler
from mlair import helpers
from mlair.helpers import remove_items
from mlair.configuration.defaults import DEFAULT_SAMPLING, DEFAULT_INTERPOLATION_LIMIT, DEFAULT_INTERPOLATION_METHOD

import inspect
from typing import Callable
import datetime as dt
from typing import Any
from functools import partial

import numpy as np
import pandas as pd
import xarray as xr


class DataHandlerMixedSamplingSingleStation(DataHandlerSingleStation):
    _requirements = remove_items(inspect.getfullargspec(DataHandlerSingleStation).args, ["self", "station"])

    def __init__(self, *args, **kwargs):
        """
        This data handler requires the kwargs sampling, interpolation_limit, and interpolation_method to be a 2D tuple
        for input and target data. If one of these kwargs is only a single argument, it will be applied to inputs and
        targets with this value. If one of these kwargs is a 2-dim tuple, the first element is applied to inputs and the
        second to targets respectively. If one of these kwargs is not provided, it is filled up with the same default
        value for inputs and targets.
        """
        self.update_kwargs("sampling", DEFAULT_SAMPLING, kwargs)
        self.update_kwargs("interpolation_limit", DEFAULT_INTERPOLATION_LIMIT, kwargs)
        self.update_kwargs("interpolation_method", DEFAULT_INTERPOLATION_METHOD, kwargs)
        super().__init__(*args, **kwargs)

    @staticmethod
    def update_kwargs(parameter_name: str, default: Any, kwargs: dict):
        """
        Update a single element of kwargs inplace to be usable for inputs and targets.

        The updated value in the kwargs dictionary is a tuple consisting on the value applicable to the inputs as first
        element and the target's value as second element: (<value_input>, <value_target>). If the value for the given
        parameter_name is already a tuple, it is checked to have exact two entries. If the paramter_name is not
        included in kwargs, the given default value is used and applied to both elements of the update tuple.

        :param parameter_name: name of the parameter that should be transformed to 2-dim
        :param default: the default value to fill if parameter is not in kwargs
        :param kwargs: the kwargs dictionary containing parameters
        """
        parameter = kwargs.get(parameter_name, default)
        if not isinstance(parameter, tuple):
            parameter = (parameter, parameter)
        assert len(parameter) == 2  # (inputs, targets)
        kwargs.update({parameter_name: parameter})

    def make_input_target(self):
        self._data = tuple(map(self.load_and_interpolate, [0, 1]))  # load input (0) and target (1) data
        self.set_inputs_and_targets()

    def load_and_interpolate(self, ind) -> [xr.DataArray, pd.DataFrame]:
        vars = [self.variables, self.target_var]
        stats_per_var = helpers.select_from_dict(self.statistics_per_var, vars[ind])
        data, self.meta = self.load_data(self.path[ind], self.station, stats_per_var, self.sampling[ind],
                                         self.station_type, self.network, self.store_data_locally, self.data_origin,
                                         self.start, self.end)
        data = self.interpolate(data, dim=self.time_dim, method=self.interpolation_method[ind],
                                limit=self.interpolation_limit[ind], sampling=self.sampling[ind])

        return data

    def set_inputs_and_targets(self):
        self.input_data = self._data[0].sel({self.target_dim: helpers.to_list(self.variables)})
        self.target_data = self._data[1].sel({self.target_dim: helpers.to_list(self.target_var)})

    def setup_data_path(self, data_path, sampling):
        """Sets two paths instead of single path. Expects sampling arg to be a list with two entries"""
        assert len(sampling) == 2
        return list(map(lambda x: super(__class__, self).setup_data_path(data_path, x), sampling))

    def _extract_lazy(self, lazy_data):
        _data, self.meta, _input_data, _target_data = lazy_data
        f_prep = partial(self._slice_prep, start=self.start, end=self.end)
        self._data = f_prep(_data[0]), f_prep(_data[1])
        self.input_data, self.target_data = list(map(f_prep, [_input_data, _target_data]))


class DataHandlerMixedSampling(DefaultDataHandler):
    """Data handler using mixed sampling for input and target."""

    data_handler = DataHandlerMixedSamplingSingleStation
    data_handler_transformation = DataHandlerMixedSamplingSingleStation
    _requirements = data_handler.requirements()


class DataHandlerMixedSamplingWithFilterSingleStation(DataHandlerMixedSamplingSingleStation,
                                                      DataHandlerFilterSingleStation):
    _requirements1 = DataHandlerFilterSingleStation.requirements()
    _requirements2 = DataHandlerMixedSamplingSingleStation.requirements()
    _requirements = list(set(_requirements1 + _requirements2))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _check_sampling(self, **kwargs):
        assert kwargs.get("sampling") == ("hourly", "daily")

    def make_input_target(self):
        """
        A FIR filter is applied on the input data that has hourly resolution. Lables Y are provided as aggregated values
        with daily resolution.
        """
        self._data = tuple(map(self.load_and_interpolate, [0, 1]))  # load input (0) and target (1) data
        self.set_inputs_and_targets()
        self.apply_filter()

    def estimate_filter_width(self):
        """Return maximum filter width."""
        raise NotImplementedError

    @staticmethod
    def _add_time_delta(date, delta):
        new_date = dt.datetime.strptime(date, "%Y-%m-%d") + dt.timedelta(hours=delta)
        return new_date.strftime("%Y-%m-%d")

    def update_start_end(self, ind):
        if ind == 0:  # for inputs
            estimated_filter_width = self.estimate_filter_width()
            start = self._add_time_delta(self.start, -estimated_filter_width)
            end = self._add_time_delta(self.end, estimated_filter_width)
        else:  # target
            start, end = self.start, self.end
        return start, end

    def load_and_interpolate(self, ind) -> [xr.DataArray, pd.DataFrame]:

        start, end = self.update_start_end(ind)
        # if ind == 0:  # for inputs
        #     estimated_filter_width = self.estimate_filter_width()
        #     start = self._add_time_delta(self.start, -estimated_filter_width)
        #     end = self._add_time_delta(self.end, estimated_filter_width)
        # else:  # target
        #     start, end = self.start, self.end

        vars = [self.variables, self.target_var]
        stats_per_var = helpers.select_from_dict(self.statistics_per_var, vars[ind])

        data, self.meta = self.load_data(self.path[ind], self.station, stats_per_var, self.sampling[ind],
                                         self.station_type, self.network, self.store_data_locally, self.data_origin,
                                         start, end)
        data = self.interpolate(data, dim=self.time_dim, method=self.interpolation_method[ind],
                                limit=self.interpolation_limit[ind], sampling=self.sampling[ind])
        return data

    def _extract_lazy(self, lazy_data):
        _data, self.meta, _input_data, _target_data = lazy_data
        start_inp, end_inp = self.update_start_end(0)
        self._data = tuple(map(lambda x: self._slice_prep(_data[x], *self.update_start_end(x)), [0, 1]))
        self.input_data = self._slice_prep(_input_data, start_inp, end_inp)
        self.target_data = self._slice_prep(_target_data, self.start, self.end)


class DataHandlerMixedSamplingWithKzFilterSingleStation(DataHandlerMixedSamplingWithFilterSingleStation,
                                                        DataHandlerKzFilterSingleStation):
    _requirements1 = DataHandlerKzFilterSingleStation.requirements()
    _requirements2 = DataHandlerMixedSamplingWithFilterSingleStation.requirements()
    _requirements = list(set(_requirements1 + _requirements2))

    def estimate_filter_width(self):
        """
        f = 0.5 / (len * sqrt(itr)) -> T = 1 / f
        :return:
        """
        return int(self.kz_filter_length[0] * np.sqrt(self.kz_filter_iter[0]) * 2)

    def _extract_lazy(self, lazy_data):
        _data, _meta, _input_data, _target_data, self.cutoff_period, self.cutoff_period_days = lazy_data
        super(__class__, self)._extract_lazy((_data, _meta, _input_data, _target_data))


class DataHandlerMixedSamplingWithKzFilter(DefaultDataHandler):
    """Data handler using mixed sampling for input and target. Inputs are temporal filtered."""

    data_handler = DataHandlerMixedSamplingWithKzFilterSingleStation
    data_handler_transformation = DataHandlerMixedSamplingWithKzFilterSingleStation
    _requirements = data_handler.requirements()


class DataHandlerMixedSamplingWithFirFilterSingleStation(DataHandlerMixedSamplingWithFilterSingleStation,
                                                         DataHandlerFirFilterSingleStation):
    _requirements1 = DataHandlerFirFilterSingleStation.requirements()
    _requirements2 = DataHandlerMixedSamplingWithFilterSingleStation.requirements()
    _requirements = list(set(_requirements1 + _requirements2))

    def estimate_filter_width(self):
        """Filter width is determined by the filter with the highest order."""
        return max(self.filter_order)

    def _extract_lazy(self, lazy_data):
        _data, _meta, _input_data, _target_data, self.fir_coeff = lazy_data
        super(__class__, self)._extract_lazy((_data, _meta, _input_data, _target_data))

    @staticmethod
    def _get_fs(**kwargs):
        """Return frequency in 1/day (not Hz)"""
        sampling = kwargs.get("sampling")[0]
        if sampling == "daily":
            return 1
        elif sampling == "hourly":
            return 24
        else:
            raise ValueError(f"Unknown sampling rate {sampling}. Only daily and hourly resolution is supported.")


class DataHandlerMixedSamplingWithFirFilter(DefaultDataHandler):
    """Data handler using mixed sampling for input and target. Inputs are temporal filtered."""

    data_handler = DataHandlerMixedSamplingWithFirFilterSingleStation
    data_handler_transformation = DataHandlerMixedSamplingWithFirFilterSingleStation
    _requirements = data_handler.requirements()


class DataHandlerMixedSamplingWithClimateFirFilterSingleStation(DataHandlerMixedSamplingWithFilterSingleStation,
                                                                DataHandlerClimateFirFilterSingleStation):
    _requirements1 = DataHandlerClimateFirFilterSingleStation.requirements()
    _requirements2 = DataHandlerMixedSamplingWithFilterSingleStation.requirements()
    _requirements = list(set(_requirements1 + _requirements2))

    def estimate_filter_width(self):
        """Filter width is determined by the filter with the highest order."""
        return max(self.filter_order)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _extract_lazy(self, lazy_data):
        _data, _meta, _input_data, _target_data, self.climate_filter_coeff, self.apriori, self.all_apriori = lazy_data
        DataHandlerMixedSamplingWithFilterSingleStation._extract_lazy(self, (_data, _meta, _input_data, _target_data))

    @staticmethod
    def _get_fs(**kwargs):
        """Return frequency in 1/day (not Hz)"""
        sampling = kwargs.get("sampling")[0]
        if sampling == "daily":
            return 1
        elif sampling == "hourly":
            return 24
        else:
            raise ValueError(f"Unknown sampling rate {sampling}. Only daily and hourly resolution is supported.")


class DataHandlerMixedSamplingWithClimateFirFilter(DefaultDataHandler):
    """Data handler using mixed sampling for input and target. Inputs are temporal filtered."""

    data_handler = DataHandlerMixedSamplingWithClimateFirFilterSingleStation
    data_handler_transformation = DataHandlerMixedSamplingWithClimateFirFilterSingleStation
    _requirements = data_handler.requirements()


class DataHandlerSeparationOfScalesSingleStation(DataHandlerMixedSamplingWithKzFilterSingleStation):
    """
    Data handler using mixed sampling for input and target. Inputs are temporal filtered and depending on the
    separation frequency of a filtered time series the time step delta for input data is adjusted (see image below).

    .. image:: ../../../../../_source/_plots/separation_of_scales.png
        :width: 400

    """

    _requirements = DataHandlerMixedSamplingWithKzFilterSingleStation.requirements()
    _hash = DataHandlerMixedSamplingWithKzFilterSingleStation._hash + ["time_delta"]

    def __init__(self, *args, time_delta=np.sqrt, **kwargs):
        assert isinstance(time_delta, Callable)
        self.time_delta = time_delta
        super().__init__(*args, **kwargs)

    def make_history_window(self, dim_name_of_inputs: str, window: int, dim_name_of_shift: str) -> None:
        """
        Create a xr.DataArray containing history data.

        Shift the data window+1 times and return a xarray which has a new dimension 'window' containing the shifted
        data. This is used to represent history in the data. Results are stored in history attribute.

        :param dim_name_of_inputs: Name of dimension which contains the input variables
        :param window: number of time steps to look back in history
                Note: window will be treated as negative value. This should be in agreement with looking back on
                a time line. Nonetheless positive values are allowed but they are converted to its negative
                expression
        :param dim_name_of_shift: Dimension along shift will be applied
        """
        window = -abs(window)
        data = self.input_data
        self.history = self.stride(data, dim_name_of_shift, window, offset=self.window_history_offset)

    def stride(self, data: xr.DataArray, dim: str, window: int, offset: int = 0) -> xr.DataArray:

        # this is just a code snippet to check the results of the kz filter
        # import matplotlib
        # matplotlib.use("TkAgg")
        # import matplotlib.pyplot as plt
        # xr.concat(res, dim="filter").sel({"variables":"temp", "Stations":"DEBW107", "datetime":"2010-01-01T00:00:00"}).plot.line(hue="filter")

        time_deltas = np.round(self.time_delta(self.cutoff_period)).astype(int)
        start, end = window, 1
        res = []
        _range = list(map(lambda x: x + offset, range(start, end)))
        window_array = self.create_index_array(self.window_dim, _range, squeeze_dim=self.target_dim)
        for delta, filter_name in zip(np.append(time_deltas, 1), data.coords["filter"]):
            res_filter = []
            data_filter = data.sel({"filter": filter_name})
            for w in _range:
                res_filter.append(data_filter.shift({dim: -(w - offset) * delta - offset}))
            res_filter = xr.concat(res_filter, dim=window_array).chunk()
            res.append(res_filter)
        res = xr.concat(res, dim="filter").compute()
        return res

    def estimate_filter_width(self):
        """
        Attention: this method returns the maximum value of
        * either estimated filter width f = 0.5 / (len * sqrt(itr)) -> T = 1 / f or
        * time delta method applied on the estimated filter width mupliplied by window_history_size
        to provide a sufficiently wide filter width.
        """
        est = self.kz_filter_length[0] * np.sqrt(self.kz_filter_iter[0]) * 2
        return int(max([self.time_delta(est) * self.window_history_size, est]))


class DataHandlerSeparationOfScales(DefaultDataHandler):
    """Data handler using mixed sampling for input and target. Inputs are temporal filtered and different time step
    sizes are applied in relation to frequencies."""

    data_handler = DataHandlerSeparationOfScalesSingleStation
    data_handler_transformation = DataHandlerSeparationOfScalesSingleStation
    _requirements = data_handler.requirements()
