__author__ = "Lukas Leufen"
__date__ = '2021-05-25'

from functools import reduce, partial

from mlair.model_modules import AbstractModelClass
from mlair.helpers import select_from_dict
from mlair.model_modules.loss import var_loss, custom_loss

import keras


class RNN(AbstractModelClass):
    """

    """

    _activation = {"relu": keras.layers.ReLU, "tanh": partial(keras.layers.Activation, "tanh"),
                   "sigmoid": partial(keras.layers.Activation, "sigmoid"),
                   "linear": partial(keras.layers.Activation, "linear"),
                   "selu": partial(keras.layers.Activation, "selu"),
                   "prelu": partial(keras.layers.PReLU, alpha_initializer=keras.initializers.constant(value=0.25)),
                   "leakyrelu": partial(keras.layers.LeakyReLU)}
    _initializer = {"tanh": "glorot_uniform", "sigmoid": "glorot_uniform", "linear": "glorot_uniform",
                    "relu": keras.initializers.he_normal(), "selu": keras.initializers.lecun_normal(),
                    "prelu": keras.initializers.he_normal()}
    _optimizer = {"adam": keras.optimizers.adam, "sgd": keras.optimizers.SGD}
    _regularizer = {"l1": keras.regularizers.l1, "l2": keras.regularizers.l2, "l1_l2": keras.regularizers.l1_l2}
    _requirements = ["lr", "beta_1", "beta_2", "epsilon", "decay", "amsgrad", "momentum", "nesterov", "l1", "l2"]
    _dropout = {"selu": keras.layers.AlphaDropout}
    _rnn = {"lstm": keras.layers.LSTM, "gru": keras.layers.GRU}

    def __init__(self, input_shape: list, output_shape: list, activation="relu", activation_output="linear",
                 optimizer="adam", n_layer=1, n_hidden=10, regularizer=None, dropout=None, layer_configuration=None,
                 batch_normalization=False, rnn_type="lstm", add_dense_layer=False, **kwargs):
        """
        Sets model and loss depending on the given arguments.

        :param input_shape: list of input shapes (expect len=1 with shape=(window_hist, station, variables))
        :param output_shape: list of output shapes (expect len=1 with shape=(window_forecast))

        Customize this RNN model via the following parameters:

        :param activation: set your desired activation function. Chose from relu, tanh, sigmoid, linear, selu, prelu,
            leakyrelu. (Default relu)
        :param activation_output: same as activation parameter but exclusively applied on output layer only. (Default
            linear)
        :param optimizer: set optimizer method. Can be either adam or sgd. (Default adam)
        :param n_layer: define number of hidden layers in the network. Given number of hidden neurons are used in each
            layer. (Default 1)
        :param n_hidden: define number of hidden units per layer. This number is used in each hidden layer. (Default 10)
        :param layer_configuration: alternative formulation of the network's architecture. This will overwrite the
            settings from n_layer and n_hidden. Provide a list where each element represent the number of units in the
            hidden layer. The number of hidden layers is equal to the total length of this list.
        :param dropout: use dropout with given rate. If no value is provided, dropout layers are not added to the
            network at all. (Default None)
        :param batch_normalization: use batch normalization layer in the network if enabled. These layers are inserted
            between the linear part of a layer (the nn part) and the non-linear part (activation function). No BN layer
            is added if set to false. (Default false)
        :param rnn_type: define which kind of recurrent network should be applied. Chose from either lstm or gru. All
            units will be of this kind. (Default lstm)
        """

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.activation = self._set_activation(activation.lower())
        self.activation_name = activation
        self.activation_output = self._set_activation(activation_output.lower())
        self.activation_output_name = activation_output
        self.optimizer = self._set_optimizer(optimizer.lower(), **kwargs)
        self.bn = batch_normalization
        self.add_dense_layer = add_dense_layer
        self.layer_configuration = (n_layer, n_hidden) if layer_configuration is None else layer_configuration
        self.RNN = self._rnn.get(rnn_type.lower())
        self._update_model_name(rnn_type)
        # self.kernel_initializer = self._initializer.get(activation, "glorot_uniform")
        # self.kernel_regularizer = self._set_regularizer(regularizer, **kwargs)
        self.dropout, self.dropout_rate = self._set_dropout(activation, dropout)

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options["loss"][0], var_loss=var_loss)

    def set_model(self):
        """
        Build the model.
        """
        if isinstance(self.layer_configuration, tuple) is True:
            n_layer, n_hidden = self.layer_configuration
            conf = [n_hidden for _ in range(n_layer)]
        else:
            assert isinstance(self.layer_configuration, list) is True
            conf = self.layer_configuration

        x_input = keras.layers.Input(shape=self._input_shape)
        x_in = keras.layers.Reshape((self._input_shape[0], reduce((lambda x, y: x * y), self._input_shape[1:])))(
            x_input)

        for layer, n_hidden in enumerate(conf):
            return_sequences = (layer < len(conf) - 1)
            x_in = self.RNN(n_hidden, return_sequences=return_sequences)(x_in)
            if self.bn is True:
                x_in = keras.layers.BatchNormalization()(x_in)
            x_in = self.activation(name=f"{self.activation_name}_{layer + 1}")(x_in)
            if self.dropout is not None:
                x_in = self.dropout(self.dropout_rate)(x_in)

        if self.add_dense_layer is True:
            x_in = keras.layers.Dense(min(self._output_shape ** 2, conf[-1]), ame=f"Dense_{len(conf) + 1}")(x_in)
            x_in = self.activation(name=f"{self.activation_name}_{len(conf) + 1}")(x_in)
        x_in = keras.layers.Dense(self._output_shape)(x_in)
        out = self.activation_output(name=f"{self.activation_output_name}_output")(x_in)
        self.model = keras.Model(inputs=x_input, outputs=[out])
        print(self.model.summary())

        # x_in = keras.layers.LSTM(32)(x_in)
        # if self.dropout is not None:
        #     x_in = self.dropout(self.dropout_rate)(x_in)
        # x_in = keras.layers.RepeatVector(self._output_shape)(x_in)
        # x_in = keras.layers.LSTM(32, return_sequences=True)(x_in)
        # if self.dropout is not None:
        #     x_in = self.dropout(self.dropout_rate)(x_in)
        # out = keras.layers.TimeDistributed(keras.layers.Dense(1))(x_in)
        # out = keras.layers.Flatten()(out)

    def _set_dropout(self, activation, dropout_rate):
        if dropout_rate is None:
            return None, None
        assert 0 <= dropout_rate < 1
        return self._dropout.get(activation, keras.layers.Dropout), dropout_rate

    def _set_activation(self, activation):
        try:
            return self._activation.get(activation.lower())
        except KeyError:
            raise AttributeError(f"Given activation {activation} is not supported in this model class.")

    def set_compile_options(self):
        self.compile_options = {"loss": [keras.losses.mean_squared_error],
                                "metrics": ["mse", "mae", var_loss]}

    def _set_optimizer(self, optimizer, **kwargs):
        try:
            opt_name = optimizer.lower()
            opt = self._optimizer.get(opt_name)
            opt_kwargs = {}
            if opt_name == "adam":
                opt_kwargs = select_from_dict(kwargs, ["lr", "beta_1", "beta_2", "epsilon", "decay", "amsgrad"])
            elif opt_name == "sgd":
                opt_kwargs = select_from_dict(kwargs, ["lr", "momentum", "decay", "nesterov"])
            return opt(**opt_kwargs)
        except KeyError:
            raise AttributeError(f"Given optimizer {optimizer} is not supported in this model class.")
    #
    # def _set_regularizer(self, regularizer, **kwargs):
    #     if regularizer is None or (isinstance(regularizer, str) and regularizer.lower() == "none"):
    #         return None
    #     try:
    #         reg_name = regularizer.lower()
    #         reg = self._regularizer.get(reg_name)
    #         reg_kwargs = {}
    #         if reg_name in ["l1", "l2"]:
    #             reg_kwargs = select_from_dict(kwargs, reg_name, remove_none=True)
    #             if reg_name in reg_kwargs:
    #                 reg_kwargs["l"] = reg_kwargs.pop(reg_name)
    #         elif reg_name == "l1_l2":
    #             reg_kwargs = select_from_dict(kwargs, ["l1", "l2"], remove_none=True)
    #         return reg(**reg_kwargs)
    #     except KeyError:
    #         raise AttributeError(f"Given regularizer {regularizer} is not supported in this model class.")
    #

    def _update_model_name(self, rnn_type):
        n_input = str(reduce(lambda x, y: x * y, self._input_shape))
        n_output = str(self._output_shape)
        self.model_name = rnn_type.upper()
        if isinstance(self.layer_configuration, tuple) and len(self.layer_configuration) == 2:
            n_layer, n_hidden = self.layer_configuration
            self.model_name += "_".join(["", n_input, *[f"{n_hidden}" for _ in range(n_layer)], n_output])
        else:
            self.model_name += "_".join(["", n_input, *[f"{n}" for n in self.layer_configuration], n_output])
